/*
 *  Stefan's Portable MakeTurn: Generate Turn File
 *
 *  (c) 1999,2002 Stefan Reuther <Streu@gmx.de>
 */
#include <cstdio>
#include <string>
#include <iomanip>
#include <cctype>
#include <fstream>
#include <iostream>
#include "turnfile.h"
#include "buffer.h"

#include "commands.def"

using namespace std;

CBuffer fizzbin;                // registration info

enum { BUFSIZE = 4096 };

int tha_ship_id;                // ship id for THost alliance hack
char tha_ship_fcode[3];         // fcode of that ship

char winplan_user_name[50], winplan_user_address[50];

extern bool winplan_turn;

/* load a file into buffer */
/* throws string error message on failure */
static void load_file(string name, CBuffer& buf) throw(string)
{
    FILE* fp = fopen(name.c_str(), "rb");
    if(!fp)
        throw string("can't open " + name);

    char buffer[BUFSIZE];
    while(1) {
        size_t n = fread(buffer, 1, BUFSIZE, fp);
        if(!n)
            break;
        buf.Append(buffer, n);
    }
    fclose(fp);
}

/* generate commands for one object */
static void generate_commands_for_obj(CTurnFile* trn,     // .TRN
                                      char* this_dat,     // .DAT record
                                      char* this_dis,     // .DIS record
                                      int id,             // object id
                                      TCommand* commands) // command defs
{
    if (!tha_ship_id && commands == ship_commands) {
        /* save ship id for THost alliance hack */
        tha_ship_id = id;
        std::memcpy(tha_ship_fcode, &this_dat[4], 3);
    }
        
    while(commands->code != -1) {
        bool thiscmd = false;
        
        /* special handling for command `build starbase' */
        int n = commands->code==34 ? 2 : commands->len;
        
        /* need this command? */
        for(int i = 0; i < n; i++)
            if(this_dat[i + commands->ofs] != this_dis[i + commands->ofs]) {
                thiscmd = true;
                break;
            }
        if (winplan_turn && commands->code == 53 && this_dat[commands->ofs] != 0)
            thiscmd = true;

        /* yes! */
        if(thiscmd) {
            trn->AddCommand(commands->code);
            trn->Data().AppendWord(id);
            trn->AddData(&this_dat[commands->ofs], commands->len);
        }
        commands++;
    }
}

/* Generate commands for dat/dis */
static void gen_commands(CTurnFile* trn,      // .TRN file
                         CBuffer& dat,        // .DAT file
                         CBuffer& dis,        // .DIS file
                         int rec_size,        // size of one record
                         int idofs,           // offset of ID no. in record
                         TCommand* commands,  // command list
                         string name)         // name of this object type
{
    int n = dat.GetWord(0);
    if(n != dis.GetWord(0))
        throw name + " data files mismatch";
    
    unsigned size = rec_size * n + 2;
    if(dat.Length() < size || dis.Length() < size)
        throw name + " data files are invalid";

    cout << setw(4) << n << " " << name << "s" << endl << flush;

    for(int i = 0; i < n; i++) {
        char* this_dat = &dat[2 + i * rec_size];
        char* this_dis = &dis[2 + i * rec_size];
        int id = dat.GetWord(2 + i * rec_size + idofs);
        if(id != dis.GetWord(2 + i * rec_size + idofs))
            throw name + " data files mismatch";
        generate_commands_for_obj(trn, this_dat, this_dis, id, commands);
    }
}

/* integer -> string */
static string itoa(int n)
{
    char buffer[20];
    sprintf(buffer, "%d", n);
    return buffer;
}

static void send_string(CTurnFile* trn, std::string str, int player)
{
    trn->AddCommand(60);
    trn->Data().AppendWord(str.length());
    trn->Data().AppendWord(player);
    trn->Data().AppendWord(player);
    for (string::size_type i = 0; i < str.length(); ++i)
        str[i] += 13;
    trn->AddData(str.data(), str.length());
}

static void chomp(string& s)
{
    string::size_type n = s.length();
    while (n && (s[n-1] == '\015' || s[n-1] == '\012'))
        --n;
    s.erase(n);
}

static void ltrim(string& s)
{
    string::size_type i = 0;
    while (i < s.length() && isspace(s[i]))
        ++i;
    if (i)
        s.erase(0, i);
}

static bool pmatch(const char* pattern, const char* tester)
{
    while (*pattern) {
        unsigned char p = *pattern++;
        if (!*tester) {
            return islower(p);  // tester ends before pattern
        } else {
            // mandatory character must match exactly
            if (tolower(p) != tolower((unsigned char) *tester))
                return false;
        }
        ++tester;
    }
    return *tester == 0;        // pattern ends - tester must end, too
}

static void send_thost_allies(CTurnFile* trn, std::string code)
{
    if (code.empty())
        return;
    if (!tha_ship_id) {
        std::cerr << "WARNING: THost alliance commands given but no ship available\n"
                     "WARNING:   the alliance commands have been ignored.\n";
        return;
    }

    /* send alliance fcodes */
    string::size_type i = 0;
    while (i < code.length()) {
        char fc[3];
        for (int j = 0; j < 3; ++j)
            if (i < code.length())
                fc[j] = code[i++];
            else
                fc[j] = ' ';
        trn->AddCommand(1);
        trn->Data().AppendWord(tha_ship_id);
        trn->AddData(fc, 3);
    }
    /* restore fcode */
    trn->AddCommand(1);
    trn->Data().AppendWord(tha_ship_id);
    trn->AddData(tha_ship_fcode, 3);
}

static void send_command_messages(CTurnFile* trn, std::string name,
                                  int player, const CBuffer& gen)
{
    ifstream file (name.c_str());
    if (!file)
        return;
    
    string line;
    string accum;
    while (getline(file, line)) {
        chomp(line);
        ltrim(line);
        if (line.empty() || line[0] == '#')
            continue;

        /* this is a command we potentially want to send */
        if (line[0] == '$') {
            /* command is to be processed by Maketurn */
            std::string arg;
            std::string::size_type i = line.find(' ');
            if (i != line.npos) {
                arg = line.substr(i);
                line.erase(i);
                ltrim(arg);
            }
            if (pmatch("$THOST-ALLIES", line.c_str())) {
                send_thost_allies(trn, arg);
            } else if (pmatch("$TIMEstamp", line.c_str())) {
                if (arg.length() != 18) {
                    cerr << "WARNING: invalid `$time' command encountered -- command has been ignored\n";
                } else if (string(gen.Ptr(), 18) != arg) {
                    cerr << "WARNING: file `" << name << "' is stale and has been ignored\n";
                    break;
                }
            } else {
                cout << "NOTE: unknown command `" << line << "' has been ignored." << endl;
            }
        } else {
            /* command is to be processed by Host, i.e. send it */
            if (accum.length() && accum.length() + line.length() > 500) {
                send_string(trn, accum, player);
                accum = line;
            } else {
                if (accum.length())
                    accum += '\015';
                accum += line;
            }
        }
    }
    if (accum.length())
        send_string(trn, accum, player);
}

/* generate turn file commands */
CTurnFile* GenerateTurnCommands(const char* dir, int player)
{
    /* initialize */
    tha_ship_id = 0;
    
    /* load GENx.DAT */
    CBuffer gen;
    try {
        load_file(dir + string("/gen") + itoa(player) + ".dat", gen);
    }
    catch(string& e) {
        return 0;
    }
    
    cout << "Generating Turn File for Player " << player << endl;
    if(gen.Length() < 155)
        throw string("gen" + itoa(player) + ".dat is invalid");

    CTurnFile* trn = new CTurnFile(player, gen.Ptr(), gen.GetWord(153));

    /* load data files */
    CBuffer pdata, pdata1;
    CBuffer bdata, bdata1;
    CBuffer sdata, sdata1;

    try {
        string p = itoa(player);
        load_file(dir + string("/pdata") + p + ".dat", pdata);
        load_file(dir + string("/pdata") + p + ".dis", pdata1);
        load_file(dir + string("/bdata") + p + ".dat", bdata);
        load_file(dir + string("/bdata") + p + ".dis", bdata1);
        load_file(dir + string("/ship") + p + ".dat", sdata);
        load_file(dir + string("/ship") + p + ".dis", sdata1);
    }
    catch(...) {
        delete trn;
        throw;
    }

    gen_commands(trn, sdata, sdata1, 107, 0, ship_commands,   "Ship");
    gen_commands(trn, pdata, pdata1,  85, 2, planet_commands, "Planet");
    gen_commands(trn, bdata, bdata1, 156, 0, base_commands,   "Starbase");

    try {
        /* send DOS Format messages */
        CBuffer mess;
        load_file(dir + string("/mess") + itoa(player) + ".dat", mess);
        int nmesg = mess.GetWord(0);
        for(int i = 0; i < nmesg; i++) {
            int head = 10 * i + 2;
            int msg  = mess.GetLong(head) - 1;
            trn->AddCommand(60);
            trn->AddData(&mess[head + 4], 6);
            trn->AddData(&mess[msg], mess.GetWord(head + 4));
        }
        cout << setw(4) << nmesg << " Messages" << endl << flush;
    }
    catch(...) { }

    /* Send command messages */
    send_command_messages(trn, dir + string("/cmd") + itoa(player) + ".txt",
                          player, gen);
    
    if(gen[141] == 13) {
        /* change password */
        trn->AddCommand(61);
        trn->Data().AppendWord(0);
        trn->AddData(&gen[143], 10);
    }
    
    return trn;
}

static void pad_space(char* ptr, size_t size)
{
    size_t i = strlen(ptr);
    while (i < size)
        ptr[i++] = ' ';
}

static bool load_reg_key(string name)
{
    try {
        CBuffer key;
        load_file(name, key);
        if (key.Length() < 154 || key.GetLong(150) != 13) {
            cout << "Winplan reg.key found but invalid (not yet activated?)\n";
            return false;
        }
        for (int i = 0; i < 50; ++i)
            winplan_user_name[i] = key[i]-13;
        for (int i = 0; i < 50; ++i)
            winplan_user_address[i] = key[i+50]-13;
        cout << "Winplan reg.key loaded.\n";
        return true;
    }
    catch(...) {
        return false;
    }
}

/* load registration info */
void LoadRegistration(const char* dir)
{
    load_file(string(dir) + "/fizz.bin", fizzbin);
    if (!load_reg_key(string(dir) + "/reg.key") && !load_reg_key(string(dir) + "/../reg.key")) {
        strcpy(winplan_user_name, "Stefan Reuther's Portable Maketurn");
        strcpy(winplan_user_address, "http://phost.de/~stefan/planets/");
        pad_space(winplan_user_name, sizeof winplan_user_name);
        pad_space(winplan_user_address, sizeof winplan_user_address);
    }
}
